#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "scp_session.h"
#include "scp_log.h"

void _KillPidWaitForTerm(int pid)
{
    int status;
    pid_t waitr;
    
    kill(pid, SIGINT);
    while (1) {
    	waitr = waitpid(pid, &status, WNOHANG);
    	SCP_DBG("pid=%d, waitpid=%d, ifexited=%d, ifsignaled=%d\n", pid, waitr, WIFEXITED(status), WIFSIGNALED(status));
    	if (-1==waitr && errno==ECHILD) {
    		break;
    	}
    	if (pid==waitr && WIFEXITED(status)) {
    		break;
    	}
    	if (pid==waitr && WIFSIGNALED(status)) {
    		break;
    	}
    	kill(pid, SIGINT);
    	sleep(1);
    }
}

static void _SigChldHndl(int pid)
{
	int status;
	
	while (waitpid(-1, &status, WNOHANG)>0);
}

static int sopcastPid = -1;
static int playerPid = -1;

static void _ParentProcessIntHndl(int sig)
{
	SCP_DBG("Got SIGTEM\n");
	if (playerPid>0)
		_KillPidWaitForTerm(playerPid);
	if (sopcastPid>0)
		_KillPidWaitForTerm(sopcastPid);
}

static int IsPortCanConnect(int sockFd, unsigned int port)
{
	int ret;
	struct sockaddr_in addr;
	
	bzero(&addr,sizeof(addr));
	addr.sin_family = AF_INET;
	inet_aton("127.0.0.1", &addr.sin_addr);
	SCP_DBG("pport=%d\n", (int)port);
	addr.sin_port = htons(port);
	
	SCP_DBG("Con\n");
	if ((ret = connect(sockFd, (struct sockaddr*)&addr, sizeof(addr)))<0 && errno==EISCONN) {
		SCP_DBG("ret =%d, errno=%d, EALREADY=%d\n Con error:%s\n", ret, errno, EALREADY, strerror(errno));
		SCP_DBG("player port avalaible now\n");
		return TRUE;
	}
	SCP_DBG("ret =%d, errno=%d, EALREADY=%d\n Con error:%s\n", ret, errno, EALREADY, strerror(errno));
	return FALSE;
}

static void _BSRespHndl(GtkDialog *dialog, gint respId, gpointer data)
{
	ScpSession* ss;
	
	ss = (ScpSession*)data;
	if (respId == GTK_RESPONSE_REJECT) {
		ScpSessionStop(ss);
		gtk_widget_hide((GtkWidget*)ss->buferringDlg);
	}
}

int ScpSessionIsBuffering(ScpSession* ss)
{
	char buf[1];
	int readed;
	
	readed = read(ss->bufferingStReadFd, buf, 1);
	if (readed<0 && errno == EAGAIN) {
		return TRUE;
	}
	close(ss->bufferingStReadFd);
	return FALSE;
}

static gboolean _CheckBuffering(gpointer data)
{
	ScpSession* ss;	
	
	ss = (ScpSession*)data;
	if (!ScpSessionIsBuffering(ss)) {
		SCP_DBG("destroy buffering dlg\n");
		gtk_progress_bar_set_fraction(ss->pBar, 1.0);
		gtk_widget_destroy((GtkWidget*)ss->buferringDlg);
		return FALSE;
	}
	gtk_progress_bar_set_fraction(ss->pBar, gtk_progress_bar_get_fraction(ss->pBar)+0.1);
	return TRUE;
}


#define SOPCAST_DATA_CONN_TIMEOUT	(50)	/* timeout = SOPCAST_DATA_CONN_TIMEOUT*2 seconds */
int ScpSessionRun(ScpSession* ss, const char* sopcastExe, int sopcastLocalPort, int sopcastPlayerPort, const char *playerExe, const char* url)
{
    int ret = -1;
    int pipeFd[2];
    int bStPipeFd[2];
    int status;
    int firstDeadPid;
    GString *localPortStr;
    GString *playerPortStr;
    GString *playerUrl;
    
    if (sopcastLocalPort<1024 || sopcastPlayerPort<1024) {
        SCP_ERR_OUT(ERR_OUT, "port should be larger than 1024\n");
    }
    signal(SIGCHLD, _SigChldHndl);
    SCP_DBG("\n");
    ss->sopcastExe = g_string_new(sopcastExe);
    ss->playerExe = g_string_new(playerExe);
    ss->url = g_string_new(url);
    SCP_DBG("\n");
    localPortStr = g_string_new("");
    g_string_printf(localPortStr, "%d", sopcastLocalPort);
    playerPortStr = g_string_new("");
    g_string_printf(playerPortStr, "%d", sopcastPlayerPort);
    playerUrl = g_string_new("");
    g_string_printf(playerUrl, "http://127.0.0.1:%d/tv.asf", sopcastPlayerPort);
    SCP_DBG("\n");
    if (pipe(pipeFd)) {
        SCP_ERR_OUT(ERR_FREE_STRINGS, "create pipe failed\n");
    }
 	if (pipe(bStPipeFd)) {
 		SCP_ERR_OUT(ERR_CLOSE_PIPE, "create pipe for buffering status failed\n");
 	}
    ss->pipeReadFd = pipeFd[0];
    ss->bufferingStReadFd = bStPipeFd[0];
    fcntl(pipeFd[0], F_SETFL, fcntl(ss->pipeReadFd, F_GETFL)|O_NONBLOCK);
    fcntl(bStPipeFd[0], F_SETFL, fcntl(ss->pipeReadFd, F_GETFL)|O_NONBLOCK);
	ss->pPid = fork();
	SCP_DBG("\n");
	if (ss->pPid == 0) {    /* child process */
		signal(SIGCHLD, SIG_DFL);
	    signal(SIGTERM, _ParentProcessIntHndl);
		setsid();
		setpgrp();
	    close(pipeFd[0]);
	    close(bStPipeFd[0]);
	    
	    sopcastPid = fork();
	    SCP_DBG("\n");
	    if (sopcastPid == 0) {  /* child process */
	        SCP_DBG("\n");
	        int stdoutFd;
	    	
	    	close(pipeFd[1]);
	    	close(bStPipeFd[1]);
	    	if((stdoutFd=open("/dev/null", O_WRONLY, 0666))>=0) {
	    		dup2(stdoutFd, 1);
	    	}
	        //execlp("eog", "eog", NULL);
	        execlp(ss->sopcastExe->str, ss->sopcastExe->str, ss->url->str, localPortStr->str, playerPortStr->str, NULL);
	    } else if (sopcastPid > 0) {    /* parent process */
	        SCP_DBG("sopcastPid=%d\n", sopcastPid);
	    } else {
	        SCP_ERR("fork for sopcast failed\n");
	        exit(-1);
	    }
	    int i = 0;
	    int sockFd;
	    sockFd = socket(AF_INET, SOCK_STREAM, 0);
	    if (sockFd<0) {
	    	_KillPidWaitForTerm(sopcastPid);
	    	exit(-1);
		}
		fcntl(sockFd, F_SETFL, fcntl(sockFd, F_GETFL)|O_NONBLOCK);
	    while (i<SOPCAST_DATA_CONN_TIMEOUT) {
	    	int retPid;
	    	retPid = waitpid(sopcastPid, &status, WNOHANG);
	    	if ((sopcastPid == retPid) 
	    			&& (WIFEXITED(status) || WIFSIGNALED(status))) {
	    		exit(-1);
	    	}else if (retPid < 0 && errno == ECHILD) {
	        	exit(-1);
	        }
	    	if (IsPortCanConnect(sockFd, (unsigned short)atoi(playerPortStr->str))) {
	    		break;
	    	}
	    	sleep(2);
	    	i++;
	    }
	    close(sockFd);
	    if (i == SOPCAST_DATA_CONN_TIMEOUT) {
	    	_KillPidWaitForTerm(sopcastPid);
	    	exit(-1);
	    }
	    close(bStPipeFd[1]);  /* Buffering finished!! */
	    //sleep(5); /* sleep for buffering, need better solution. */
	    playerPid = fork();
	    SCP_DBG("\n");
	    if (playerPid == 0) {  /* child process */
	        SCP_DBG("\n");
	        //execlp("gnome-session-properties", "gnome-session-properties", NULL);
	        execlp(ss->playerExe->str, ss->playerExe->str, playerUrl->str, NULL);
	    } else if (playerPid > 0) {    /* parent process */
	    	SCP_DBG("playerPid=%d\n", playerPid);
	        SCP_DBG("\n");
	    } else {
	        SCP_ERR("fork for player failed\n");
            _KillPidWaitForTerm(sopcastPid);
	        exit(errno);
	    }
	    while (1) {
	        firstDeadPid = wait(&status);
	        SCP_DBG("firstDeadPid=%d, errno=%d, EINTR=%d\n", firstDeadPid, errno, EINTR);
	        if (firstDeadPid == playerPid && (WIFEXITED(status)
	        		|| WIFSIGNALED(status))) {
	            _KillPidWaitForTerm(sopcastPid);
	            exit(0);
	        } else if (firstDeadPid == sopcastPid && (WIFEXITED(status)
	        		|| WIFSIGNALED(status))) {
	            _KillPidWaitForTerm(playerPid);
	            exit(-1);
	        } else if (firstDeadPid < 0 && errno == ECHILD) {
	        	exit(-1);
	        } else {
	                _KillPidWaitForTerm(playerPid);
	                _KillPidWaitForTerm(sopcastPid);
	                exit(errno);
	        }
	    }
	    exit(0);
	} else if (ss->pPid > 0) {  /* parent process */
		
	    SCP_DBG("ppid=%d\n", ss->pPid);
	    ss->buferringDlg = (GtkDialog*)gtk_dialog_new_with_buttons(_("Buffering..."), NULL, 0, 
		         GTK_STOCK_CANCEL,
		         GTK_RESPONSE_REJECT,
		         NULL);
    	g_signal_connect(ss->buferringDlg, "response", G_CALLBACK(_BSRespHndl), ss);
    	ss->pBar = (GtkProgressBar*)gtk_progress_bar_new();
    	gtk_widget_set_hexpand((GtkWidget*)ss->pBar, TRUE);
    	gtk_progress_bar_set_fraction(ss->pBar, 0.0);
    	gtk_progress_bar_set_text(ss->pBar, _("Buffering, Please wait ..."));
    	gtk_progress_bar_set_show_text(ss->pBar, TRUE);
    	gtk_widget_show((GtkWidget*)ss->pBar);
    	gtk_container_set_border_width((GtkContainer*)gtk_dialog_get_content_area(ss->buferringDlg), 8);
    	gtk_container_add((GtkContainer*)gtk_dialog_get_content_area(ss->buferringDlg), (GtkWidget*)ss->pBar);
    	gtk_widget_show((GtkWidget*)ss->buferringDlg);
    	close(pipeFd[1]);
		close(bStPipeFd[1]);
		g_timeout_add(1000, _CheckBuffering, ss);
	} else {
	    ret = ss->pPid;
	    SCP_ERR_OUT(ERR_CLOSE_B_PIPE, "fork failed\n");
	}

	g_string_free(localPortStr, TRUE);
	g_string_free(playerPortStr, TRUE);
	g_string_free(playerUrl, TRUE);
	g_string_free(ss->sopcastExe, TRUE);
	g_string_free(ss->playerExe, TRUE);
    return 0;
    
ERR_CLOSE_B_PIPE:
    close(bStPipeFd[1]);
    close(bStPipeFd[0]);
ERR_CLOSE_PIPE:
    close(pipeFd[1]);
    close(pipeFd[0]);
ERR_FREE_STRINGS:
	g_string_free(localPortStr, TRUE);
	g_string_free(playerPortStr, TRUE);
	g_string_free(playerUrl, TRUE);
	g_string_free(ss->sopcastExe, TRUE);
	g_string_free(ss->playerExe, TRUE);
ERR_OUT:
    return ret;
}

int ScpSessionStop(ScpSession* ss)
{
	char buf[1];
	int readed;
	
	readed = read(ss->pipeReadFd, buf, 1);
	SCP_DBG("readed=%d\n", readed);
	if (readed== 0) {
		return -1;
	}
	SCP_DBG("kill ppid=%d\n", ss->pPid);
	kill(ss->pPid, SIGTERM);
	while (read(ss->pipeReadFd, buf, 1) != 0){sleep(1);}
	return 0;
}

#if 0//def _DEBUG_

int main(int argc, char** argv)
{
    ScpSession ss;
    SCP_DBG("\n");
    ScpSessionRun(&ss, "sp-sc", 3908, 8908, "totem", "sop://218.106.52.252:3912/23597");
    pause();
}
#endif


